from typing import Any, List
import lightning.pytorch as L
import torch
import torch.nn as nn
from torch.optim import AdamW
from transformers.optimization import (
    Adafactor,
    get_cosine_schedule_with_warmup,
    get_cosine_with_hard_restarts_schedule_with_warmup,
    get_constant_schedule_with_warmup,
    get_linear_schedule_with_warmup,
    get_polynomial_decay_schedule_with_warmup,
)


# update this and the import above to support new schedulers from transformers.optimization
arg_to_scheduler = {
    "linear": get_linear_schedule_with_warmup,
    "cosine": get_cosine_schedule_with_warmup,
    "cosine_w_restarts": get_cosine_with_hard_restarts_schedule_with_warmup,
    "polynomial": get_polynomial_decay_schedule_with_warmup,
    "constant": get_constant_schedule_with_warmup,  # not supported for now
}



class BaseLitModel(L.LightningModule):
    def __init__(self, config, tokenizer):
        super().__init__()
        self.config = config
        self.tokenizer = tokenizer
        self.stage = 'train'
        if "dropout" not in config:
            self.dropout = None
        else:
            self.dropout = nn.Dropout(config.dropout)
        self.validation_step_outputs = []

    def training_step(self, batch: Any, batch_idx: int):
        outputs = self(**batch)
        if outputs['loss'] != outputs['loss']:
            raise Exception("Loss为Nan，请先检查数据正确性！")
        self.log("loss", outputs['loss'], prog_bar=True, logger=True, sync_dist=True, on_step=True,
                         on_epoch=True, rank_zero_only=True)
        for key in outputs.keys():
            if '_loss' in key:
                key_loss = round(float(torch.clamp(outputs[key].cpu(), max=99, min=0)), 3)
                self.log("train/" + key, key_loss, prog_bar=False, logger=True, sync_dist=True, on_step=True,
                         on_epoch=True, rank_zero_only=True)
        if 'lm_loss' in outputs:
            ppl = round(float(torch.clamp(torch.exp(outputs['lm_loss']).cpu(), max=99, min=0)), 3)
            self.log("ppl", ppl, prog_bar=False, logger=True, sync_dist=True, rank_zero_only=True)
        
        self.log("lr", round(self.opt.param_groups[0]['lr'], 9), prog_bar=True, logger=True, on_step=True,
                 rank_zero_only=True, sync_dist=True)
        self.log("current_epoch", float(self.current_epoch))
        self.log("progress", self.global_step / self.total_steps())
        return {
            'loss': outputs['loss'],
        }

    def validation_step(self, batch: Any, batch_idx: int):
        outputs = self(**batch)
        self.log("val/step_loss", outputs['loss'], prog_bar=False)
        self.validation_step_outputs.append(outputs['loss'])
        return {"val_step_loss": outputs['loss']}

    def on_validation_epoch_end(self):
        mean_loss = torch.stack([x for x in self.validation_step_outputs]).mean().item()
        log = {
            'val_loss': mean_loss,
        }
        self.log('val_loss', mean_loss)
        self.validation_step_outputs.clear()
        return log

    def prepare_other_features_for_generation(self, batch):
        ignore_keys = ['input_ids', 'labels', 'decoder_input_ids']
        other_features = dict()
        for key in batch.keys():
            if key not in ignore_keys:
                try:
                    other_features[key] = torch.LongTensor(batch[key]).to(self.device)
                except ValueError as e:
                    other_features[key] = batch[key]
        return other_features


    def get_lr_scheduler(self):
        lr_scheduler = self.config.get("lr_scheduler", "constant")
        get_schedule_func = arg_to_scheduler[lr_scheduler]
        total_steps = self.total_steps()
        if self.config.get("warmup_ratio", 0.1) > 0:
            warmup_steps = self.config.get("warmup_ratio", 0.1) * total_steps
        else:
            warmup_steps = self.config.get("warmup_steps", 100)

        if lr_scheduler != "constant":
            scheduler = get_schedule_func(self.opt, num_warmup_steps=warmup_steps, num_training_steps=total_steps)
        else:
            scheduler = get_schedule_func(self.opt, num_warmup_steps=warmup_steps)

        scheduler = {"scheduler": scheduler, "interval": "step", "frequency": 1}
        return scheduler

    def configure_optimizers(self):
        """Prepare optimizer and schedule (linear warmup and decay)"""
        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in self.named_parameters() if not any(nd in n for nd in no_decay)],
                "weight_decay": self.config.get("weight_decay", 0.0),
            },
            {
                "params": [p for n, p in self.named_parameters() if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            },
        ]
        if self.config.get("adafactor", None):
            optimizer = Adafactor(
                optimizer_grouped_parameters,
                lr=self.config.get("lr", 1e-5),
                scale_parameter=False,
                relative_step=False,
            )

        else:
            optimizer = AdamW(
                optimizer_grouped_parameters,
                lr=self.config.get("lr", 1e-5),
                eps=self.config.get("adam_epsilon", 0.1),
            )
        self.opt = optimizer
        scheduler = self.get_lr_scheduler()
        return [optimizer] , [scheduler]

    def total_steps(self) -> int:
        return self.trainer.estimated_stepping_batches



from lazydl.utils.result import Result
from lazydl.utils.log import Logger
from lazydl.utils.catch_error import print_error_info
from datasets import Dataset
import pandas as pd
from evaluate import load
from statistics import mean
import re
import string
from collections import Counter
from bert_score import score
from nltk import word_tokenize
from nltk.translate.bleu_score import corpus_bleu, sentence_bleu, SmoothingFunction
from nltk.translate.meteor_score import meteor_score
from nltk.util import ngrams
from rouge import Rouge
from sacrebleu.metrics import BLEU, CHRF
import sacrebleu
import spacy
from rich.console import Console
from rich.table import Table
from typing import Union

log = Logger(__name__) 

def distinct_ngram(candidates, n=2):
    """Return basic ngram statistics, as well as a dict of all ngrams and their freqsuencies."""
    ngram_freqs = {}  # ngrams with frequencies
    ngram_len = 0  # total number of ngrams
    for candidate in candidates:
        for ngram in ngrams(word_tokenize(candidate), n):
            ngram_freqs[ngram] = ngram_freqs.get(ngram, 0) + 1
            ngram_len += 1
    # number of unique ngrams
    uniq_ngrams = len([val for val in ngram_freqs.values() if val == 1])
    distinct_ngram = len(ngram_freqs) / ngram_len if ngram_len > 0 else 0
    return round(distinct_ngram, 4)

def compute_meteor(references, candidates):
    score_list = []
    for i in range(len(candidates)):
        if type(references[i]) is list:
            ref_list = references[i]
        else:
            ref_list = [references[i]]
        ref = [r.split(" ") for r in ref_list]
        cand = candidates[i].split(" ")
        score = meteor_score(ref, cand)
        score_list.append(score)
    return round(mean(score_list), 4)

def compute_sent_bleu(references, candidates):
    bleu1 = 0.0
    bleu2 = 0.0
    bleu3 = 0.0
    bleu4 = 0.0
    ref_list, dec_list = [], []
    for i in range(len(candidates)):
        dec_list.append(word_tokenize(candidates[i]))
        if type(references[i]) is list:
            tmp = []
            for ref in references[i]:
                tmp.append(word_tokenize(ref))
            ref_list.append(tmp)
        else:
            ref_list.append([word_tokenize(references[i])])

    for example_id, (label, pred) in enumerate(zip(ref_list, dec_list)):
        bleu1 += sentence_bleu(
            label,
            pred,
            weights=[1, 0, 0, 0],
            smoothing_function=SmoothingFunction().method3,
        )
        bleu2 += sentence_bleu(
            label,
            pred,
            weights=[0.5, 0.5, 0, 0],
            smoothing_function=SmoothingFunction().method3,
        )
        bleu3 += sentence_bleu(
            label,
            pred,
            weights=[1 / 3, 1 / 3, 1 / 3, 0],
            smoothing_function=SmoothingFunction().method3,
        )
        bleu4 += sentence_bleu(
            label,
            pred,
            weights=[0.25, 0.25, 0.25, 0.25],
            smoothing_function=SmoothingFunction().method3,
        )
    bleu1 = bleu1 / len(ref_list)
    bleu2 = bleu2 / len(ref_list)
    bleu3 = bleu3 / len(ref_list)
    bleu4 = bleu4 / len(ref_list)
    return (
        round(bleu1 * 100, 4),
        round(bleu2 * 100, 4),
        round(bleu3 * 100, 4),
        round(bleu4 * 100, 4),
    )
    
def compute_rouge(references, candidates):
    rouge = Rouge()
    scores = rouge.get_scores(candidates, references)
    rouge_1 = [score["rouge-1"]["f"] * 100 for score in scores]
    rouge_2 = [score["rouge-2"]["f"] * 100 for score in scores]
    rouge_l = [score["rouge-l"]["f"] * 100 for score in scores]
    return (
        round(mean(rouge_1), 5),
        round(mean(rouge_2), 5),
        round(mean(rouge_l), 5),
    )

def compute_chrf(references, candidates):
    chrf = CHRF(word_order=2)
    return round(chrf.corpus_score(candidates, [references]).score, 4)


def compute_sacre_corpus_bleu(references, candidates):
    bleu = BLEU()
    return round(bleu.corpus_score(candidates, [references]).score, 4)


def compute_sacre_sent_bleu(references, candidates):
    bleu = []
    for i, a_gold in enumerate(references):
        bleu.append(sacrebleu.corpus_bleu([candidates[i]], [[a_gold]]).score)
    return round(mean(bleu), 4)

def get_tokens(text, nlp):
    doc = nlp(text)
    tokens = [tok.text.lower()
              for tok in doc if not tok.is_stop and not tok.is_punct]
    return tokens


def get_bert_score(references, candidates, device):
    # scorer = BertScorer()
    # scorer.init_scorer(lang='en', num_layers=8, rescale_with_baseline=True)
    # scores = scorer.get_score(target, generated)[-1]
    scores = score(
        candidates,
        references,
        lang="en",
        verbose=False,
        rescale_with_baseline=True,
        device=device,
    )[-1].numpy()
    return round(mean(list(scores)), 4)

def compute_corpus_bleu(references, candidates):
    bleu1 = 0.0
    bleu2 = 0.0
    bleu3 = 0.0
    bleu4 = 0.0
    ref_list, dec_list = [], []
    for i in range(len(candidates)):
        dec_list.append(word_tokenize(candidates[i]))
        if type(references[i]) is list:
            tmp = []
            for ref in references[i]:
                tmp.append(word_tokenize(ref))
            ref_list.append(tmp)
        else:
            ref_list.append([word_tokenize(references[i])])
    bleu1 = corpus_bleu(ref_list, dec_list, weights=(1, 0, 0, 0))
    bleu2 = corpus_bleu(ref_list, dec_list, weights=(0, 1, 0, 0))
    bleu3 = corpus_bleu(ref_list, dec_list, weights=(0, 0, 1, 0))
    bleu4 = corpus_bleu(ref_list, dec_list, weights=(0, 0, 0, 1))
    return (
        round(bleu1 * 100, 4),
        round(bleu2 * 100, 4),
        round(bleu3 * 100, 4),
        round(bleu4 * 100, 4),
    )

def compute_f1(references, candidates, nlp=None):
    """
    This function is copied from: https://github.com/orhonovich/q-squared/blob/main/pipeline/score.py
    """
    f1_list = []
    for i, a_gold in enumerate(references):
        a_pred = candidates[i]
        if a_pred == "":
            f1_list.append(0)
            continue
        if nlp:
            gold_toks = get_tokens(a_gold, nlp)
            pred_toks = get_tokens(a_pred, nlp)
        else:
            gold_toks = clean_text(a_gold).split()
            pred_toks = clean_text(a_pred).split()
        common = Counter(gold_toks) & Counter(pred_toks)
        num_same = sum(common.values())
        if num_same == 0:
            f1_list.append(0)
            continue
        precision = 1.0 * num_same / len(pred_toks)
        recall = 1.0 * num_same / len(gold_toks)
        f1 = (2 * precision * recall) / (precision + recall)
        f1_list.append(f1)
    return round(sum(f1_list) / len(f1_list) * 100, 2)


def clean_text(text):
    text = text.lower()
    text = text.translate(str.maketrans("", "", string.punctuation))
    text = re.sub(r"\b(a|an|the|in|our)\b", " ", text)
    return re.sub(" +", " ", text).strip()




def get_eval_metrics(outputs: Union[Dataset, pd.DataFrame, dict] = None, 
                     metrics: list = [], 
                     device: int = 0):
    """计算评价指标

    Args:
        outputs (Union[Dataset, pd.DataFrame, dict], optional): 模型的输出
        metrics (list, optional): 需要评价的指标. Defaults to [].
        device (int, optional): 如果涉及 GPU 计算，想要使用的 GPU 序号，应该为单 GPU. Defaults to 0.
        
    """
    if not isinstance(outputs, Dataset):
        try:
            if isinstance(outputs, pd.DataFrame):
                outputs = Dataset.from_pandas(outputs)
            elif isinstance(outputs, dict):
                outputs = Dataset.from_dict(outputs)
            else:
                raise ValueError()
        except Exception as e:
            raise ValueError("评价指标计算的输入必须是 Dataset, pd.DataFrame 或者 dict类型")
    test_result = Result()
    if "generated_seqs" in outputs.column_names:
        generated_seqs = outputs["generated_seqs"]
    else:
        generated_seqs = outputs["generated"]
    if "reference" in outputs.column_names:
        reference = outputs["reference"]
    else:
        return test_result

    # ---------------------------------------------------------------------------- #
    #                         计算 Meteor                                     
    # ---------------------------------------------------------------------------- #
    if "meteor" in metrics:
        log.info("计算 Meteor ing...")
        meteor_score = compute_meteor(reference, generated_seqs)
        test_result.add(meteor=meteor_score)
        log.info(f"Meteor = {str(meteor_score)}")

    # ---------------------------------------------------------------------------- #
    #                         计算 chrf                                     
    # ---------------------------------------------------------------------------- #
    if "hf_chrf" in metrics:
        log.info("计算 chrF ing...")
        try:
            chrf_computer = load("chrf")
            chrf = chrf_computer.compute(
                predictions=generated_seqs, references=reference)['score']
            hf_chrf = round(chrf, 4)
        except Exception as e:
            log.error("计算 hf_chrf 失败")
            print_error_info(e)
            hf_chrf = 0
        test_result.add(chrf=hf_chrf)
        log.info(f"chrf = {str(hf_chrf)}")

    if "chrf" in metrics:
        try:
            chrf = compute_chrf(reference, generated_seqs)  # 偏低3个点
        except Exception as e:
            log.error("计算 chrf 失败")
            print_error_info(e)
            chrf = 0
        test_result.add(chrf=chrf)
        log.info(f"chrf = {str(chrf)}")

    # ---------------------------------------------------------------------------- #
    #                                  计算F1                                     
    #     https://github.com/orhonovich/q-squared/blob/main/pipeline/score.py
    # ---------------------------------------------------------------------------- #
    if "f1_space_split" in metrics:
        log.info("计算 f1_space_split ing...")
        f1_reference = outputs["f1_reference"]
        f1 = compute_f1(f1_reference, generated_seqs)
        test_result.add(f1_space_split=f1)
        log.info(f"f1_space_split = {str(f1)}")

    if "f1_nlp_split" in metrics:
        log.info("计算 f1_nlp_split ing...")
        nlp = spacy.load("en_core_web_sm")
        f1_reference = outputs["f1_reference"]
        f1 = compute_f1(f1_reference, generated_seqs, nlp)
        test_result.add(f1_nlp_split=f1)
        log.info(f"f1_nlp_split = {str(f1)}")

    # ---------------------------------------------------------------------------- #
    #                         计算 google_bleu                                     
    # ---------------------------------------------------------------------------- #
    if "hf_google_bleu" in metrics:
        log.info("计算 hf_google_bleu ing...")
        try:
            google_bleu = load("hf_google_bleu")
            google_bleu_score = google_bleu.compute(
                predictions=generated_seqs, references=reference
            )["hf_google_bleu"]
            google_bleu_score = google_bleu_score * 100
        except Exception as e:
            log.error("计算 hf_google_bleu 失败")
            print_error_info(e)
            google_bleu_score = 9999
        google_bleu_score = round(google_bleu_score, 4)
        test_result.add(google_bleu=google_bleu_score)
        log.info(f"hf_google_bleu = {str(google_bleu_score)}")

    # ---------------------------------------------------------------------------- #
    #                         计算 sacrebleu                                     
    # ---------------------------------------------------------------------------- #
    if "hf_sacrebleu" in metrics:
        log.info("计算 HuggingFcae Sacre BLEU ing...")
        try:
            sacrebleu_computer = load("sacrebleu")
            bleu = sacrebleu_computer.compute(
                predictions=generated_seqs, references=reference)["score"]
            bleu = round(bleu, 4)
        except Exception as e:
            log.error("计算 hf_sacrebleu 失败")
            print_error_info(e)
            bleu = 0
        test_result.add(sacrebleu_huggingface=bleu)
        log.info(f"hf_sacrebleu = {str(bleu)}")

    if "sacrebleu_sent" in metrics:
        log.info("计算 Sacre BLEU sent level ing...")
        try:
            bleu = compute_sacre_sent_bleu(
                candidates=generated_seqs, references=reference)
        except Exception as e:
            log.error("计算 sacrebleu_sent 失败")
            print_error_info(e)
            bleu = 0
        test_result.add(sacrebleu_sent=bleu)
        log.info(f"sacrebleu_sent = {str(bleu)}")

    if "sacrebleu_corpus" in metrics:
        log.info("计算 Sacre BLEU corpus level ing...")
        try:
            bleu = compute_sacre_corpus_bleu(
                candidates=generated_seqs, references=reference)
        except Exception as e:
            log.error("计算 sacrebleu_corpus 失败")
            print_error_info(e)
            bleu = 0
        test_result.add(sacrebleu_corpus=bleu)
        log.info(f"sacrebleu_corpus = {str(bleu)}")

    # ---------------------------------------------------------------------------- #
    #                         sent_bleu                                     
    # ---------------------------------------------------------------------------- #
    if "sent_bleu" in metrics:
        bleu1, bleu2, bleu3, bleu4 = compute_sent_bleu(
            reference, generated_seqs)
        mean_sent_bleu = round((bleu1 + bleu2 + bleu3 + bleu4) / 4, 4)
        test_result.add(
            sent_bleu1=bleu1,
            sent_bleu2=bleu2,
            sent_bleu3=bleu3,
            sent_bleu4=bleu4,
            mean_sent_bleu=mean_sent_bleu,
        )
        log.info(f"sent_bleu1 = {str(bleu1)}")
        log.info(f"sent_bleu2 = {str(bleu2)}")
        log.info(f"sent_bleu3 = {str(bleu3)}")
        log.info(f"sent_bleu4 = {str(bleu4)}")
        log.info(f"mean_sent_bleu = {str(mean_sent_bleu)}")

    # ---------------------------------------------------------------------------- #
    #                         corpus_bleu                                     
    # ---------------------------------------------------------------------------- #
    if "corpus_bleu" in metrics:
        bleu1, bleu2, bleu3, bleu4 = compute_corpus_bleu(
            reference, generated_seqs)
        mean_corpus_bleu = round((bleu1 + bleu2 + bleu3 + bleu4) / 4, 4)
        test_result.add(
            corpus_bleu1=bleu1,
            corpus_bleu2=bleu2,
            corpus_bleu3=bleu3,
            corpus_bleu4=bleu4,
            mean_corpus_bleu=mean_corpus_bleu,
        )
        log.info(f"corpus_bleu1 = {str(bleu1)}")
        log.info(f"corpus_bleu2 = {str(bleu2)}")
        log.info(f"corpus_bleu3 = {str(bleu3)}")
        log.info(f"corpus_bleu4 = {str(bleu4)}")
        log.info(f"mean_corpus_bleu = {str(mean_corpus_bleu)}")

    # ---------------------------------------------------------------------------- #
    #                         计算 Dist                                     
    # ---------------------------------------------------------------------------- #
    if "dist" in metrics:
        log.info("计算 Dist ing...")
        dist1 = distinct_ngram(generated_seqs, n=1)
        dist2 = distinct_ngram(generated_seqs, n=2)
        test_result.add(
            dist1=dist1,
            dist2=dist2,
        )
        log.info(f"dist1 = {str(dist1)}")
        log.info(f"dist2 = {str(dist2)}")

    # ---------------------------------------------------------------------------- #
    #                         计算 ROUGE                                     
    # ---------------------------------------------------------------------------- #
    if "hf_rouge" in metrics:
        log.info("计算 hf_rouge ing...")
        try:
            rouge = load('rouge')

            rouge_results = rouge.compute(predictions=generated_seqs,
                                          references=reference)
            rouge_1 = round(rouge_results['rouge1'], 5)
            rouge_2 = round(rouge_results['rouge2'], 5)
            rouge_L = round(rouge_results['rougeL'], 5)
            rouge_Lsum = round(rouge_results['rougeLsum'], 5)
            test_result.add(rouge_1=rouge_1)
            test_result.add(rouge_2=rouge_2)
            test_result.add(rouge_L=rouge_L)
            test_result.add(rouge_Lsum=rouge_Lsum)
            log.info(f"hf_rouge_1 = {str(rouge_1)}")
            log.info(f"hf_hf_rouge_2 = {str(rouge_2)}")
            log.info(f"hf_rouge_L = {str(rouge_L)}")
            log.info(f"hf_rouge_Lsum = {str(rouge_Lsum)}")

        except Exception as e:
            print_error_info(e)
            log.info("hf_rouge 无法计算，Reference可能为空，请检查生成数据！")

    if "rouge" in metrics:
        log.info("计算 ROUGE ing...")
        try:
            rouge_1, rouge_2, rouge_l = compute_rouge(reference, generated_seqs)
            test_result.add(rouge_1=rouge_1)
            test_result.add(rouge_2=rouge_2)
            test_result.add(rouge_l=rouge_l)
            log.info(f"rouge_1 = {str(rouge_1)}")
            log.info(f"rouge_2 = {str(rouge_2)}")
            log.info(f"rouge_L = {str(rouge_l)}")

        except Exception as e:
            print_error_info(e)
            log.info("Rouge 无法计算，Reference可能为空，请检查生成数据！")

    # ---------------------------------------------------------------------------- #
    #                         计算 Bert Score                                     
    # ---------------------------------------------------------------------------- #
    if "hf_bert_score" in metrics:
        log.info("计算 hf_bert_score ing...")
        bert_score_reference = outputs["bert_score_reference"]
        try:
            bertscore = load("bertscore")
            bert_score = mean(bertscore.compute(predictions=generated_seqs,
                                                references=bert_score_reference, lang="en", rescale_with_baseline=True)['f1'])
            bert_score = round(bert_score, 4)
            test_result.add(bert_score=bert_score)
            log.info(f"hf_bert_score = {str(bert_score)}")
        except ValueError:
            log.info("hf_bert_score 无法计算，Reference可能为空，请检查生成数据！")

    if "bert_score" in metrics:
        log.info("计算 Bert score ing...")
        bert_score_reference = outputs["bert_score_reference"]
        try:
            bert_score = get_bert_score(bert_score_reference, generated_seqs, device)
            test_result.add(bert_score=bert_score)
            log.info(f"bert_score = {str(bert_score)}")
        except ValueError:
            log.info("bert_score 无法计算，Reference可能为空，请检查生成数据！")

    # ---------------------------------------------------------------------------- #
    #                         计算 PPL                                     
    # ---------------------------------------------------------------------------- #
    if "hf_ppl" in metrics:
        log.info("计算 hf_ppl ing...")
        perplexity = load("perplexity", module_type="metric")
        try:
            ppl = perplexity.compute(predictions=generated_seqs, model_id="gpt2")[
                "mean_perplexity"
            ]
        except Exception as e:
            log.error("计算 PPL 失败")
            print_error_info(e)
            ppl = 9999
        ppl = round(ppl, 4)
        test_result.add(ppl=ppl)
        log.info(f"hf_ppl = {str(ppl)}")

    # ---------------------------------------------------------------------------- #
    #                         计算 q_squared                                     
    # ---------------------------------------------------------------------------- #
    if "q_squared" in metrics:
        from q2.run_pipeline import q2_step_one
        from q2.run_nli import q2_step_two
        log.info("计算 q_squared ing...")

        try:
            df, steps_df = q2_step_one(infile=pd.DataFrame(outputs)[["generated_seqs", "reference", "q2_reference"]].rename(columns={
                "generated_seqs": "response",
                "reference": "gold",
                "q2_reference": "knowledge"
            }), gen_method="beam", q_per_cand="single", personal="remove", device=device)
            q_squared_nli, q_squared_f1 = q2_step_two(infile=steps_df, device=device)
            test_result.add(q_squared_nli=round(q_squared_nli, 4))
            test_result.add(q_squared_f1=round(q_squared_f1, 4))
            log.info(f"q_squared_nli = {str(q_squared_nli)}")
            log.info(f"q_squared_f1 = {str(q_squared_f1)}")
        except Exception as e:
            print_error_info(e)
            log.info("q_squared 无法计算，请检查数据！")

    console = Console(color_system="256", style="cyan")
    table = Table(style="cyan", show_footer=False, title="[bold green]Evaluation results")
    table.add_column("Metric", justify="right", style="magenta")
    table.add_column("Score:yum:", justify="left", style="magenta")
    for k, v in test_result.items():
        table.add_row(k, str(v))
    console.print(table)
    return test_result


import json
import os
import datetime
import time
from redis import Redis
from lazydl.utils.log import Logger

log = Logger(__name__) 

class RedisClient:
    def __init__(self):
        self.client = Redis(
            host="127.0.0.1",
            port=6379,
            decode_responses=True,
            charset="UTF-8",
            encoding="UTF-8",
        )

    def get_self_occupied_gpus(self, only_gpus=True):
        """
        获取自己已经占用的Gpu序号
        """
        self_occupied_gpus = self.client.hgetall("self_occupied_gpus")
        if only_gpus:
            all_gpus = []
            for task in self_occupied_gpus.values():
                gpus = [
                    int(device) for device in str(json.loads(task)["cuda_devices"]).split(",")
                ]
                all_gpus.extend(gpus)
            return set(all_gpus)
        return [json.loads(g) for g in self_occupied_gpus.values()]

    def join_wait_queue(self, processing_unit_type, processing_unit, units_count, memo):
        """
        加入等待队列
        """
        curr_time = datetime.datetime.now()
        creat_time = datetime.datetime.strftime(curr_time, "%Y-%m-%d %H:%M:%S")
        task_id = (
            str(os.getpid())
            + "*"
            + str(int(time.mktime(time.strptime(creat_time, "%Y-%m-%d %H:%M:%S"))))
        )
        content = {
            "processing_unit_type": processing_unit_type,
            "processing_unit": processing_unit,
            "units_count": units_count,
            "create_time": creat_time,
            "update_time": creat_time,
            "system_pid": os.getpid(),
            "task_id": task_id,
            "task_desc": memo,
        }
        wait_num = len(self.client.lrange("wait_queue", 0, -1))
        self.client.rpush("wait_queue", json.dumps(content))
        if wait_num == 0:
            log.info(f"正在排队中！ 目前排第一位哦！")
        else:
            log.info(f"正在排队中！ 前方还有 {wait_num} 个训练任务！")
        return task_id, wait_num

    def is_my_turn(self, task_id):
        """
        排队这么长时间，是否轮到我了？
        """
        curr_task = json.loads(self.client.lrange("wait_queue", 0, -1)[0])
        return curr_task["task_id"] == task_id

    def update_queue(self, task_id):
        """
        更新等待队列
        """
        task = json.loads(self.client.lrange("wait_queue", 0, -1)[0])
        if task["task_id"] != task_id:
            # 登记异常信息
            log.info("当前训练任务并不排在队列第一位，请检查Redis数据正确性！")
        curr_time = datetime.datetime.now()
        update_time = datetime.datetime.strftime(
            curr_time, "%Y-%m-%d %H:%M:%S")
        task["update_time"] = update_time
        self.client.lset("wait_queue", 0, json.dumps(task))

    def pop_wait_queue(self, task_id):
        """
        弹出当前排位第一的训练任务
        """
        task = json.loads(self.client.lrange("wait_queue", 0, -1)[0])
        if task["task_id"] != task_id:
            # 登记异常信息
            log.info("当前训练任务并不排在队列第一位，请检查Redis数据正确性！")
        next_task = self.client.lpop("wait_queue")
        return next_task

    def register_gpus(self, task_id, processing_unit_type, processing_unit, units_count, memo):
        """
        将当前训练任务登记到GPU占用信息中
        """
        curr_time = datetime.datetime.now()
        creat_time = datetime.datetime.strftime(curr_time, "%Y-%m-%d %H:%M:%S")
        if not task_id:
            task_id = (
                str(os.getpid())
                + "*"
                + str(int(time.mktime(time.strptime(creat_time, "%Y-%m-%d %H:%M:%S"))))
            )
        else:
            task_id = task_id

        content = {
            "processing_unit_type": processing_unit_type,
            "processing_unit": processing_unit,
            "units_count": units_count,
            "create_time": creat_time,
            "update_time": creat_time,
            "system_pid": os.getpid(),
            "task_id": task_id,
            "task_desc": memo,
        }
        self.client.hset("self_occupied_gpus", task_id, json.dumps(content))
        log.info("成功登记Gpu使用信息到Redis服务器！")
        return task_id

    def deregister_gpus(self, task_id):
        """
        删除当前训练任务的占用信息
        """
        task = self.client.hget("self_occupied_gpus", task_id)
        if task:
            self.client.hdel("self_occupied_gpus", task_id)
            log.info("成功删除Redis服务器上的Gpu使用信息！")
        else:
            log.info("无法找到当前训练任务在Redis服务器上的Gpu使用信息！或许可以考虑检查一下Redis的数据 🤔")

    def register_process(self, task_id, processing_unit_type, processing_unit, units_count, memo):
        """
        将当前训练任务登记到进程信息中
        """
        curr_time = datetime.datetime.now()
        creat_time = datetime.datetime.strftime(curr_time, "%Y-%m-%d %H:%M:%S")
        if not task_id:
            task_id = (
                str(os.getpid())
                + "*"
                + str(int(time.mktime(time.strptime(creat_time, "%Y-%m-%d %H:%M:%S"))))
            )
        else:
            task_id = task_id

        content = {
            "processing_unit_type": processing_unit_type,
            "processing_unit": processing_unit,
            "units_count": units_count,
            "create_time": creat_time,
            "update_time": creat_time,
            "system_pid": os.getpid(),
            "task_id": task_id,
            "task_desc": memo,
        }
        self.client.hset("running_processes", task_id, json.dumps(content))
        log.info("成功登记进程使用信息到Redis服务器！")
        return task_id

    def deregister_process(self, task_id):
        """
        删除当前训练任务的信息
        """
        task = self.client.hget("running_processes", task_id)
        if task:
            self.client.hdel("running_processes", task_id)
            log.info("成功删除Redis服务器上的进程使用信息！")
        else:
            log.info("无法找到当前训练任务在Redis服务器上的进程使用信息！或许可以考虑检查一下Redis的数据 🤔")



import copy

class Result(dict):
    """返回结果基类

    Args:
        dict (_type_): 初始化字典
    """
    def __getattr__(self, name):
        return self[name]

    def __init__(self, *args, **kwargs):
        super(Result, self).__init__()
        for arg in args:
            for key, value in arg.items():
                self[key] = value
        self.add(**kwargs)

    # 序列化时调用
    def __getstate__(self):
        return None

    def add(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    def delete(self, keys):
        for k in keys:
            self.pop(k)

    def merge(self, merge_dict):
        if not isinstance(merge_dict, Result):
            raise TypeError("不支持的合并类型")
        for k, v in merge_dict.items():
            if k in ["msg", "status"] or k in self:
                continue
            self[k] = v

    def merge_or_update(self, merge_dict):
        if not isinstance(merge_dict, Result) and not isinstance(merge_dict, dict):
            raise TypeError("不支持的合并类型")
        for k, v in merge_dict.items():
            if k in ["msg", "status"]:
                continue
            self[k] = v

    @staticmethod
    def create_error_msg_result(msg="Error Result", **kwargs):
        result = Result()
        result["msg"] = msg
        result["status"] = False
        result.add(**kwargs)
        return result

    def get(self, name, other=None):
        if name is None:
            return list(self.values())
        elif isinstance(name, str):
            return self[name] if name in self else other
        elif isinstance(name, list):
            values = [self[n] for n in name]
            return values
        else:
            return self.create_error_msg_result(msg=f"Key值类型{type(name)}不支持")

    def print(self, name=None):
        print("  =====" + self["msg"] + "=====")
        values = self.get(name)
        if name is None:
            name = list(self.keys())
        for i, k in enumerate(name):
            v = values[i]
            print(f"  {k}:    {v}")
        print("  =====" + self["msg"] + "=====")

    def flatten_to_print(self):
        value_str = ""
        keys = self.keys()
        for i, k in enumerate(keys):
            v = self[k]
            value_str = value_str + k + " : " + str(v) + "\n\n"
        return value_str

    def aprintend_values(self, next_dict):
        if not isinstance(next_dict, Result) and not isinstance(next_dict, dict):
            raise TypeError("不支持的合并类型")
        for key in next_dict.keys():
            if key not in self.keys():
                self[key] = []

            self[key].aprintend(next_dict[key]) if isinstance(self[key], list) else [
                self[key]
            ].aprintend(next_dict[key])

    def str(self, key_name, default_value=""):
        return self.get(key_name, default_value)

    def bool(self, key_name, default_value=False):
        return self.get(key_name, default_value)

    def int(self, key_name, default_value=0):
        return self.get(key_name, default_value)

    def float(self, key_name, default_value=0.0):
        return self.get(key_name, default_value)

    def list(self, key_name, default_value=[]):
        return self.get(key_name, default_value)

    def dict(self, key_name, default_value={}):
        return self.get(key_name, default_value)

    def set(self, key_name, value):
        self[key_name] = value

    def set_with_dict(self, dict_value):
        for key, value in dict_value.items():
            if "." in key:
                key_list = key.split(".")
                self[key_list[0]][key_list[1]] = value
            else:
                self[key] = value

    def __deepcopy__(self, memo=None, _nil=[]):
        if memo is None:
            memo = {}
        d = id(self)
        y = memo.get(d, _nil)
        if y is not _nil:
            return y

        dict = Result()
        memo[d] = id(dict)
        for key in self.keys():
            dict.__setattr__(
                copy.deepcopy(key, memo), copy.deepcopy(
                    self.__getattr__(key), memo)
            )
        return dict

    def copy(self):
        return super().copy()


from lazydl.utils.log import Logger
import os
import pickle
import json
import jsonlines
import torch
import pandas as pd

log = Logger(__name__) 


def load_in(path, data_name=""):
    """读取文件，根据文件后缀名自动选择读取方法
        目前支持保存类型有：‘pkl’、‘txt’、‘pt’、‘json’, 'jsonl'、'csv'

    Args:
        data_name: str, 打印提示时需要，便于控制台查看保存的文件是什么文件, 默认为空

    Returns:
        data：Object
    """
    if not os.path.exists(path):
        log.info(f"文件路径似乎并不存在....")
        raise FileNotFoundError(path)
    log.info(f"正在加载文件 {data_name} from {path}")
    if ".pkl" in path:
        with open(path, "rb") as f:
            data = pickle.load(f)
    elif ".json" in path:
        with open(path, "r") as f:
            # 读取json数据
            data = json.load(f)
    elif ".jsonl" in path:
        data = []
        with open(path, "rb") as f:
            for item in jsonlines.Reader(f):  # 每一行读取后都是一个json，可以按照key去取对应的值
                data.append(item)
    elif ".pt" in path:
        data = torch.load(path)
    elif ".txt" in path:
        data = []
        for item in open(path):
            item = item.replace("\n", "").strip()
            if item == "":
                continue
            data.append(item)
    elif ".csv" in path:
        data = pd.read_csv(path)
    log.info(f"成功加载 {data_name}!")
    return data


def save_as(data, save_path, file_format="pt", data_name="", protocol=4):
    """将参数中的文件对象保存为指定格式格式文件
        目前支持保存类型有：‘pkl’、‘txt’、‘pt’、‘json’, 'jsonl'
        默认为‘pt’

    Args:
        data: obj, 要保存的文件对象
        save_path: str, 文件的保存路径，应当包含文件名
        data_name: str, 打印提示时需要，便于控制台查看保存的文件是什么文件, 默认为空
        protocol: int, 当文件特别大的时候，需要将此参数调到4以上, 默认为4
        file_format: str, 要保存的文件类型，支持‘pkl’、‘txt’、‘pt’、‘json’、‘jsonl’

    Returns:
        None
    """
    parent_path = "/".join(save_path.split("/")[:-1])
    if not os.path.exists(parent_path):
        log.info(f"保存路径的父文件夹（{parent_path}）不存在，将自动创建....")
        os.makedirs(parent_path)
    save_path = save_path + f".{file_format}"
    log.info(f"正在保存文件 {data_name} 到 {save_path}")
    if file_format == "pkl":
        with open(save_path, "wb") as f:
            pickle.dump(data, f, protocol=protocol)
    elif file_format == "txt":
        if not isinstance(data, list):
            data = [data]
        with open(save_path, "w") as f:
            for line in data:
                f.write(str(line) + "\n")
    elif file_format == "json":
        with open(save_path, "w") as f:
            json.dump(data, f)
    elif file_format == "jsonl":
        with jsonlines.open(save_path, mode="w") as writer:
            writer.write_all(data)
    elif file_format == "pt":
        torch.save(data, save_path)
    else:
        raise Exception(f"请添加针对{file_format}类型文件的保存方法！")
    log.info(f"保存 {data_name} 成功!")
    return None



from lazydl.utils.log import Logger
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
import importlib

logger = Logger(__name__)


class BaseDataset(Dataset):
    def __init__(self, file, tokenizer, max_seq_length):
        self.tokenizer = tokenizer
        self.bos_token_id = tokenizer.bos_token_id
        self.eos_token_id = tokenizer.eos_token_id
        self.max_seq_length = max_seq_length
        logger.info('Loading data: {}'.format(file))
        with open(file, 'r', encoding='utf8') as f:
            data_list = f.readlines()
        logger.info("there are {} data in dataset".format(len(data_list)))
        self.data_list = data_list

    def __len__(self):
        return len(self.data_list)



def rename_columns(dataset, column_map):
    for old_name, new_name in column_map.items():
        dataset = dataset.rename_column(old_name, new_name)
    return dataset



def load_data(config, tokenizer, return_dataloader=True, stage="train"):
    
    dataset_module = get_class(config.dataset_module_file)
    
    if stage == "train":
        data_file = config.train_data_file
    elif stage == "val":
        data_file = config.val_data_file
    elif stage == "test":
        data_file = config.test_data_file
    else:
        raise ValueError("Invalid stage")
    
    dataset = dataset_module(data_file, tokenizer, config.max_seq_length)
        
    if return_dataloader:
        data_collator = get_class(config.dataset_collator_function_file)
        if stage == "train":
            batch_size = config.get("train_batch_size", 1)
        elif stage == "val":
            batch_size = config.get("val_batch_size", 1)
        elif stage == "test":
            batch_size = config.get("test_batch_size", 1)
        else:
            raise ValueError("Invalid stage")
        return DataLoader(dataset, collate_fn=data_collator(tokenizer, config.max_seq_length), batch_size=batch_size, shuffle=True, num_workers=config.get("dataloader_num_workers", 1))
    return dataset
    
    

def get_class(module_file):
    module_path = 'projects.modules.' + ".".join(module_file.split('.')[:-1])
    class_name = module_file.split('.')[-1]
        
    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError as r:
        raise ValueError(f"Please add a class named like this: {module_path}.{class_name}.")
    except Exception as r:
        raise Exception('未知错误 %s' % r)
    module_class = getattr(module, class_name)
    return module_class

